from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, Dict, Iterable, Type, Optional
from pathlib import Path

import tomlkit
from pkg_resources import EntryPoint, iter_entry_points

from vulcan import Vulcan


@dataclass
class PluginRunner:
    vulcan: Vulcan
    plugin_configs: Dict[str, Any] = field(init=False)

    def __post_init__(self) -> None:
        try:
            pyproject = self.vulcan.source_path / 'pyproject.toml'
            self.plugin_configs = tomlkit.loads(pyproject.read_text()  # type: ignore
                                                )['tool']['vulcan']['plugin']
        except KeyError:
            self.plugin_configs = {}

    def get_pre_entrypoints(self) -> Iterable[EntryPoint]:
        return iter_entry_points('vulcan.pre_build')

    def get_post_entrypoints(self) -> Iterable[EntryPoint]:
        return iter_entry_points('vulcan.post_build')

    def __enter__(self) -> 'PluginRunner':
        if not self.vulcan.plugins:
            return self
        for ep in self.get_pre_entrypoints():
            if ep.name not in self.vulcan.plugins:
                continue
            print(f"Running pre_build plugin {ep}")
            ep.load()(self.plugin_configs.get(ep.name))
        return self

    def __exit__(self,
                 exc_type: Type[BaseException] = None,
                 exc_value: BaseException = None,
                 tb: TracebackType = None) -> None:
        if exc_type is not None:
            # if the build process raises an error, don't bother with the plugins.
            return None
        # NOT implementing post_build plugins yet, until I see a motivating example
        return None


def test_plugin(config: Optional[Dict[str, str]]) -> None:
    assert config is not None
    assert config['foobar'] == 'barfoo'
    (Path(config['module_dir']) / 'example.no-hash.py').write_text("Text!")



import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, cast

import setuptools
import tomlkit
from editables import EditableProject  # type: ignore
from ppsetuptools.ppsetuptools import _parse_kwargs  # type: ignore

from vulcan import Vulcan, flatten_reqs
from vulcan.plugins import PluginRunner

version: Callable[[str], str]
if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version


def _filter_nones(vals_dict: Dict[str, Optional[Any]]) -> Dict[str, Any]:
    return {k: v for k, v in vals_dict.items() if v is not None}


def setup(**kwargs: Any) -> Any:

    with open('pyproject.toml', 'r') as pptoml:
        pyproject_data = cast(Dict[str, Any], tomlkit.parse(pptoml.read()))

    if 'project' in pyproject_data:
        if 'dependencies' in pyproject_data['project']:
            raise RuntimeError("May not use [project]:dependencies key with vulcan")
        if 'optional-dependencies' in pyproject_data['project']:
            raise RuntimeError("May not use [project]:optional-dependencies key with vulcan")
        parsed_kwargs = _parse_kwargs(pyproject_data['project'], '.')
        parsed_kwargs.update(_filter_nones(kwargs))
        parsed_kwargs = _filter_nones(parsed_kwargs)
        # ppsetuptools doesn't handle entry points correctly
        if 'scripts' in parsed_kwargs:
            if 'entry_points' not in parsed_kwargs:
                parsed_kwargs['entry_points'] = {}
            parsed_kwargs['entry_points']['console_scripts'] = parsed_kwargs['scripts']
            del parsed_kwargs['scripts']
        if 'gui-scripts' in parsed_kwargs:
            parsed_kwargs['entry_points']['gui_scripts'] = parsed_kwargs['gui-scripts']
            del parsed_kwargs['gui-scripts']
        if 'entry_points' in parsed_kwargs:
            for ep_group in list(parsed_kwargs['entry_points']):
                parsed_kwargs['entry_points'][ep_group] = [
                    f'{k}={v}' for k, v in parsed_kwargs['entry_points'][ep_group].items()]
        return setuptools.setup(**parsed_kwargs)
    else:
        return setuptools.setup(**_filter_nones(kwargs))


__all__ = ['build_wheel',
           'build_sdist']


@contextmanager
def patch_argv(argv: List[str]) -> Generator[None, None, None]:
    old_argv = sys.argv[:]
    sys.argv = [sys.argv[0]] + argv
    yield
    sys.argv = old_argv


def build(outdir: str, config_settings: Dict[str, str] = None) -> str:
    config = Vulcan.from_source(Path().absolute())
    options: Dict[str, Any] = {}
    if config.packages:
        options['packages'] = config.packages
    if config.version:
        options['version'] = config.version

    if config.no_lock or (config_settings and config_settings.get('no-lock') == 'true'):
        options['install_requires'] = flatten_reqs(config.configured_dependencies)
        options['extras_require'] = config.configured_extras
    else:
        options['install_requires'] = config.dependencies
        options['extras_require'] = config.extras

    # https://setuptools.readthedocs.io/en/latest/userguide/keywords.html
    # https://docs.python.org/3/distutils/apiref.html
    with PluginRunner(config):
        dist = setup(**options, include_package_data=True)
    rel_dist = Path(dist.dist_files[0][-1])
    shutil.move(str(rel_dist), Path(outdir) / rel_dist.name)
    return rel_dist.name


def build_wheel(wheel_directory: str, config_settings: Dict[str, str] = None,
                metadata_directory: str = None) -> str:
    with patch_argv(['bdist_wheel']):
        return build(wheel_directory, config_settings)


def build_sdist(sdist_directory: str,
                config_settings: Dict[str, str] = None,
                ) -> str:
    with patch_argv(['sdist']):
        return build(sdist_directory, config_settings)


def get_virtualenv_python() -> Path:
    virtual_env = os.environ.get('VIRTUAL_ENV')
    if virtual_env is None:
        raise RuntimeError("No virtualenv active")
    if sys.platform == 'win32':
        # sigh
        return Path(virtual_env, 'Scripts', 'python')
    else:
        # if this isn't in an else,
        # mypy complains on windows that it is unreachable
        return Path(virtual_env, 'bin', 'python')


# tox requires these two for some reason :(
def get_requires_for_build_sdist(config_settings: Dict[str, str] = None) -> List[str]:
    return []


def get_requires_for_build_wheel(config_settings: Dict[str, str] = None) -> List[str]:
    return []


def get_pip_version(python_callable: Path) -> Optional[Tuple[int, ...]]:
    out = subprocess.check_output([str(python_callable), '-m', 'pip', '--version'], encoding='utf-8')
    m = re.search(r'pip (\d+\.\d+(\.\d+)?)', out)
    if not m:
        return None
    return tuple((int(n) for n in m.group(1).split('.')))


@contextmanager
def maybe_gen_setuppy(venv: Path, config: Vulcan) -> Generator[None, None, None]:
    pip_version = get_pip_version(venv)
    if pip_version is None or pip_version < (21, 3):
        print(f"pip version {pip_version} does not support editable installs for PEP517 projects,"
              " falling back to generated setup.py")
        options: Dict[str, Any] = {}
        if config.packages:
            options['packages'] = config.packages
        if config.version:
            options['version'] = config.version

        if config.no_lock:
            options['install_requires'] = flatten_reqs(config.configured_dependencies)
            options['extras_require'] = config.configured_extras
        else:
            options['install_requires'] = config.dependencies
            options['extras_require'] = config.extras

        setup = Path('setup.py')
        if setup.exists():
            exit('may not use vulcan develop when setup.py is present')
        try:
            with tempfile.NamedTemporaryFile(suffix='.json', mode="w+", delete=False) as mdata_file:
                try:
                    mdata_file.write(json.dumps(options))
                    mdata_file.flush()
                    mdata_file.close()
                    with setup.open('w+') as setup_file:
                        setup_file.write(f"""\
from vulcan.build_backend import setup
import json, pathlib
setup(**json.load(pathlib.Path(r'{mdata_file.name}').open()))
""")
                        yield
                finally:
                    os.unlink(mdata_file.name)
        finally:
            setup.unlink()
    else:
        yield


def install_develop() -> None:
    config = Vulcan.from_source(Path().absolute())

    try:
        virtual_env = get_virtualenv_python()
    except RuntimeError:
        exit('may not use vulcan develop outside of a virtualenv')

    with maybe_gen_setuppy(virtual_env, config):
        path = str(Path().absolute())
        if config.configured_extras:
            path = f'{path}[{",".join(config.configured_extras)}]'
        subprocess.check_call([
            str(virtual_env), '-m', 'pip', 'install', '-e', path])


# pep660 functions
def unpack(whl: Path) -> Path:
    with tempfile.TemporaryDirectory() as tmp:
        subprocess.check_output(f'wheel unpack {whl} -d {tmp}'.split())
        unpacked = list(Path(tmp).glob('*'))
        assert len(unpacked) == 1
        shutil.copytree(unpacked[0], whl.parent / unpacked[0].name)
        return whl.parent / unpacked[0].name


def pack(unpacked_wheel: Path) -> Path:
    with tempfile.TemporaryDirectory() as tmp:
        subprocess.check_output(f'wheel pack {unpacked_wheel} -d {tmp}'.split())
        packed = list(Path(tmp).glob('*.whl'))
        assert len(packed) == 1
        shutil.copy(packed[0], unpacked_wheel.parent)
        return unpacked_wheel.parent / packed[0].name


def add_requirement(unpacked_whl_dir: Path, req: str) -> None:
    metadata = next(unpacked_whl_dir.glob('*.dist-info')) / 'METADATA'  # is mandatory
    with metadata.open() as f:
        metadata_lines = list(f)
    i = 0
    for i, line in enumerate(metadata_lines):
        if not (line.strip() and not line.startswith('Requires-Dist: ')):
            # find the start of the requires-dist, or the end of the metadata keys
            break
    metadata_lines.insert(i, f'Requires-Dist: {req}\n')
    metadata.write_text(''.join(metadata_lines))


def make_editable(whl: Path) -> None:
    config = Vulcan.from_source(Path().absolute())
    unpacked_whl_dir = unpack(whl)
    add_requirement(unpacked_whl_dir, f"editables (~={version('editables')})")
    project = EditableProject(config.name, Path().absolute())
    for name, content in project.files():
        (unpacked_whl_dir / name).write_text(content)

    assert whl == pack(unpacked_whl_dir), 'pre-wheel and post-wheel should be the same path'
    shutil.rmtree(unpacked_whl_dir)


def build_editable(wheel_directory: str, config_settings: Dict[str, str] = None,
                   metadata_directory: str = None) -> str:
    whl_path = Path(wheel_directory) / build_wheel(wheel_directory, config_settings, metadata_directory)
    make_editable(whl_path)
    return whl_path.name



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, cast

import tomlkit
from typing_extensions import TypedDict


def find_version_file(source_dir: Path) -> Optional[Path]:
    try:
        return next(source_dir.rglob('VERSION'))
    except StopIteration:
        return None


class VulcanConfigError(Exception):
    pass


class VersionDict(TypedDict, total=False):
    version: str
    extras: List[str]


VersionSpecs = Mapping[str, Union[str, VersionDict]]


def flatten_reqs(versions: VersionSpecs) -> List[str]:
    return [to_pep508(lib, req) for lib, req in versions.items()]


@dataclass
class ShivOpts:
    bin_name: str
    console_script: Optional[str] = None
    entry_point: Optional[str] = None
    interpreter: Optional[str] = None
    with_extras: Optional[List[str]] = None
    extra_args: str = ''


class _ContainerStub(tomlkit.container.Container):
    def get(self, v: str, default: Any = None) -> Any:
        ...


def list_or_none(val: Any) -> Optional[List[str]]:
    return [str(v) for v in val] if val is not None else None


def str_or_none(val: Any) -> Optional[str]:
    return str(val) if val is not None else None


def dict_or_none(val: Any) -> Optional[Dict[str, Any]]:
    return {str(k): v for k, v in val.items()} if val is not None else None


@dataclass
class Vulcan:
    name: str
    version: Optional[str]
    packages: Optional[List[str]]
    source_path: Path
    plugins: Optional[List[str]]
    shiv_options: List[ShivOpts]
    lockfile: Path
    dependencies: Optional[List[str]]
    configured_dependencies: VersionSpecs
    extras: Optional[Dict[str, List[str]]]
    configured_extras: Dict[str, List[str]]
    no_lock: bool = False
    python_lock_with: Optional[str] = None

    @classmethod
    def from_source(cls, source_path: Path, fail_on_missing_lock: bool = True) -> 'Vulcan':
        with open(source_path / 'pyproject.toml') as f:
            all_config = tomlkit.loads(f.read())
            name = str(all_config['project']['name'])  # type: ignore
            config = all_config['tool']['vulcan']  # type: ignore
            config = cast(_ContainerStub, config)
        version_file = find_version_file(source_path)
        version = version_file.read_text().strip() if version_file is not None else config.get('version')
        lockfile = source_path / config.get('lockfile', 'vulcan.lock')

        no_lock = config.get('no-lock', False)
        install_requires: Optional[List[str]] = []
        extras_require: Optional[Dict[str, List[str]]] = {}
        if not no_lock:
            try:
                install_requires, extras_require = get_requires(lockfile)
            except FileNotFoundError:
                if fail_on_missing_lock:
                    raise
                install_requires = None
                extras_require = None

        python_lock_with = config.get('python-lock-with')

        shiv_ops = []
        shiv_config = config.get('shiv', [])
        for conf in shiv_config:
            shiv_ops.append(ShivOpts(
                bin_name=str(conf.get('bin_name')),
                console_script=str_or_none(conf.get('console_script')),
                entry_point=str_or_none(conf.get('entry_point')),
                interpreter=str_or_none(conf.get('interpreter')),
                with_extras=[str(e) for e in conf.get('with_extras', [])],
                extra_args=str(conf.get('extra_args', '')),
            ))

        print(f"Setting version to {version}")
        return cls(name=name,
                   version=version, source_path=source_path, plugins=list_or_none(config.get('plugins')),
                   packages=list_or_none(config.get("packages")),
                   lockfile=lockfile, shiv_options=shiv_ops,
                   dependencies=install_requires,
                   configured_dependencies=config.get('dependencies', {}),
                   extras=extras_require,
                   configured_extras=config.get('extras', {}),
                   no_lock=no_lock, python_lock_with=python_lock_with)


def get_requires(lockfile: Path) -> Tuple[List[str], Dict[str, List[str]]]:
    if not lockfile.exists():
        raise FileNotFoundError(f"Expected lockfile {lockfile}, does not exist")
    with lockfile.open() as f:
        content = cast(_ContainerStub, tomlkit.loads(f.read()))
    return (list(content['install_requires']),  # type: ignore
            {k: list(v) for k, v in content['extras_require'].items()})  # type: ignore


def to_pep508(lib: str, req: Union[str, VersionDict]) -> str:
    if not isinstance(req, (str, dict)):
        raise VulcanConfigError(f"Invalid requirement {req} -- must be a dict or a string")
    if isinstance(req, str):
        # e.g. "example_lib", "~=1.2.3" -> "example_lib~=1.2.3"
        return f'{lib}{req}'
    try:
        extras = f'[{",".join(req["extras"])}]' if 'extras' in req else ''
        # "example_lib", {"version": "~=1.2.3", "extras"=["networkx","git"]}
        #                           -> "example_lib[networkx,git]~=1.2.3"
        # "example_lib", {"version": "~=1.2.3"} -> "example_lib~=1.2.3"
        return f'{lib}{extras}{req["version"]}'
    except KeyError as e:
        raise VulcanConfigError(f'invalid requirement {lib} ({req}) -- {e}') from e



import asyncio
import shlex
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from os import PathLike
from types import SimpleNamespace
from typing import Dict, Generator, List, Union
from venv import EnvBuilder

from pkg_resources import Requirement


@contextmanager
def create_venv(python_version: str = None) -> Generator['VulcanEnvBuilder', None, None]:
    with tempfile.TemporaryDirectory(prefix='vulcan-build-') as tempdir:
        builder = VulcanEnvBuilder(with_pip=True, python_version=python_version)
        builder.create(tempdir)
        yield builder


def get_executable(version: str) -> str:
    if sys.platform == 'win32':
        which = 'where'
    else:
        which = 'which'
    return subprocess.check_output(
        [which, f'python{version}'], encoding='utf-8', stderr=subprocess.PIPE).strip()


@contextmanager
def patch_executable(python_version: str = None) -> Generator[None, None, None]:
    if python_version is None:
        yield
    else:
        try:
            if sys.version_info >= (3, 8):
                # for some reason mypy thinks this is incorrect for 3.8
                # even though it is definitely correct and I've read the code showing it
                # So this whole file disables the "warn_unused_ignore" config due to these 2 lines :(
                old_exe = sys._base_executable  # type: ignore
                sys._base_executable = get_executable(python_version)  # type: ignore
            else:
                old_exe = sys.executable
                sys.executable = get_executable(python_version)
            yield
        except subprocess.CalledProcessError as e:
            print(f"Command '{' '.join(shlex.quote(a) for a in e.cmd)}' failed with exit code {e.returncode}")
            print(e.stderr)
            exit(1)
        finally:
            if sys.version_info >= (3, 8):
                sys._base_executable = old_exe  # type: ignore
            else:
                sys.executable = old_exe


class VulcanEnvBuilder(EnvBuilder):

    def __init__(self, system_site_packages: bool = False, clear: bool = False,
                 symlinks: bool = False, upgrade: bool = False, with_pip: bool = False, prompt: str = None,
                 python_version: str = None):
        self.context: SimpleNamespace
        super().__init__(system_site_packages=system_site_packages, clear=clear, symlinks=symlinks,
                         upgrade=upgrade, with_pip=with_pip, prompt=prompt)
        self._executable_python_version = python_version

    def ensure_directories(self,
                           env_dir: Union[str, bytes, 'PathLike[str]', 'PathLike[bytes]']
                           ) -> SimpleNamespace:
        with patch_executable(self._executable_python_version):
            self.context = super().ensure_directories(env_dir)
        return self.context

    def _setup_pip(self, context: SimpleNamespace) -> None:
        super()._setup_pip(context)
        cmd = [context.env_exe, '-Im', 'pip', 'install', '--upgrade', 'pip']
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    async def install(self, deps_dir: Union[str, bytes, 'PathLike[str]', 'PathLike[bytes]'],
                      requirements: List[str]
                      ) -> None:
        # install Isolated with module pip using pep517
        if not requirements:
            return
        cmd = [
            self.context.env_exe,
            '-Im',
            'pip',
            'install',
            '--no-cache-dir',
            '--use-pep517',
            '--target',
            str(deps_dir)] + requirements
        proc = await asyncio.create_subprocess_exec(*cmd, stderr=asyncio.subprocess.PIPE,
                                                    stdout=asyncio.subprocess.PIPE)
        out, err = await proc.communicate()
        assert proc.returncode is not None
        if proc.returncode != 0:
            raise subprocess.CalledProcessError(returncode=proc.returncode, cmd=cmd, output=out, stderr=err)

    async def freeze(self, deps_dir: Union[str, bytes, 'PathLike[str]', 'PathLike[bytes]']
                     ) -> Dict[str, Requirement]:
        # list with the requirements.txt format only libraries installed in specifically this venv
        cmd = [self.context.env_exe, '-Im', 'pip', 'list', '--format=freeze', '--path', str(deps_dir)]

        frozen = await asyncio.create_subprocess_exec(*cmd, stderr=asyncio.subprocess.PIPE,
                                                      stdout=asyncio.subprocess.PIPE)

        out, err = await frozen.communicate()
        assert frozen.returncode is not None
        if frozen.returncode != 0:
            raise subprocess.CalledProcessError(returncode=frozen.returncode, cmd=cmd, output=out, stderr=err)
        reqs = [Requirement.parse(line) for line in out.decode().split('\n') if line]
        return {req.name: req for req in reqs}  # type: ignore
